/**********************************************************************
* 
* regionlock.h
*
* This file is part of MitoGen
* 
* Copyright (C) 2013-2017 -- Centre for Biomedical Image Analysis (CBIA)
* http://cbia.fi.muni.cz/
* 
* MitoGen is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* MitoGen is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with MitoGen. If not, see <http://www.gnu.org/licenses/>.
* 
* Author: David Wiesner
* 
* Description: Definition of class 'SceneLocks' that handles
* region-based locks in the multi-threaded implementation of MitoGen.
*
***********************************************************************/


#ifndef MITOGEN_REGIONLOCK_H
#define MITOGEN_REGIONLOCK_H


#include <list>
#include <condition_variable>
#include <i3d/image3d.h>

#include "../macros.h"

template <class MV, class PV> class Cell;

extern thread_local size_t thrID;


enum lock_type
{
	timePoint_lock,
	sphere_lock
};


enum scene_type
{
	all,
	masks,
	nuclei,
	phantoms,
	territories	
};
// function returning the string containing scene type (used in printLockInfo())
std::string GetSceneTypeStr(const scene_type &sceneType);


class CustomLock;
typedef std::list<CustomLock*>::iterator lock_handle;
	
	
/**
 * \ingroup toolbox
 *
 * Base class from which all custom locks are derived.
 *
 */
class CustomLock
{
public:

	CustomLock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const scene_type sceneType):
		threadID(thrID),
		fromTimePoint(fromTimePoint),
		toTimePoint(toTimePoint),
		sceneType(sceneType) {}
	
	virtual ~CustomLock() {}

	virtual lock_type getLockType() const = 0;
	
	scene_type getSceneType() const
	{ return sceneType; }

	size_t getThreadID() const
	{ return threadID; }
	
	size_t getFromTimePoint() const
	{ return fromTimePoint; }
	
	size_t getToTimePoint() const
	{ return toTimePoint; }
	
	// wrapper for printLockInfo
	void printLockInfo() const
	{
		const std::chrono::milliseconds timeToAcquireLock(0);
		printLockInfo(timeToAcquireLock);
	}
	
	virtual void printLockInfo(const std::chrono::milliseconds timeToAcquireLock) const = 0;
	
protected:
	
	const size_t threadID;
	const size_t fromTimePoint;
	const size_t toTimePoint;
	const scene_type sceneType;
	
};


/**
 * \ingroup toolbox
 *
 * Class defining the time-point lock.
 *
 */
class TimePointLock: public CustomLock
{
public:
	
	TimePointLock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const scene_type sceneType):
		CustomLock(fromTimePoint, toTimePoint, sceneType) {}
	
	lock_type getLockType() const
	{ return timePoint_lock; }
	
	void printLockInfo(const std::chrono::milliseconds timeToAcquireLock) const
	{
		REPORT("TIME lock"
			<< ", waiting time: " << timeToAcquireLock.count() << "ms"
			<< ", timePoint: " << fromTimePoint << "-" << toTimePoint
			<< ", sceneType: " << GetSceneTypeStr(sceneType));
	}
};


/**
 * \ingroup toolbox
 *
 * Class defining the sphere lock.
 *
 */
class SphereLock: public CustomLock
{
public:

	// sphere-shaped lock constructor
	// radius of sphere around each cell is 2.5*cellDiameter / 2
	template <class MV, class PV> 
	SphereLock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const Cell<MV,PV>* const cellPtr,
		const scene_type sceneType):
		CustomLock(fromTimePoint, toTimePoint, sceneType),
		sphereRadius(cellPtr->GetCellDiameter() * 1.25f),
		sphereCentre(cellPtr->GetCellCentrePosition()),
		cellID(cellPtr->GetID()) {}

	// a copy constructor with own time-interval
	SphereLock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const SphereLock& lock):
		CustomLock(fromTimePoint, toTimePoint, lock.getSceneType()),
		sphereRadius(lock.getSphereRadius()),
		sphereCentre(lock.getSphereCentre()),
		cellID(lock.getCellID()) {}

	lock_type getLockType() const
	{ return sphere_lock; }
	
	float getSphereRadius() const
	{ return sphereRadius; }
	void setSphereRadius(const float rad)
	{ sphereRadius = rad; }
	
	i3d::Vector3d<float> getSphereCentre() const
	{ return sphereCentre; }
	void updateSphereCentre(const i3d::Vector3d<float>& newCentre)
	{ sphereCentre = newCentre; }
	
	size_t getCellID() const
	{ return cellID; }
	
	void printLockInfo(const std::chrono::milliseconds timeToAcquireLock) const
	{
		REPORT("SPHERE lock"
			<< ", waiting time: " << timeToAcquireLock.count() << "ms"
			<< ", timePoint: " << fromTimePoint << "-" << toTimePoint
			<< ", sceneType: " << GetSceneTypeStr(sceneType)
			<< ", centre: [" << sphereCentre.x << ", "
			<< sphereCentre.y << ", " << sphereCentre.z
			<< "], radius: " << sphereRadius);		
	}
	
private:
	
	// radius of the sphere (in microns)
	float sphereRadius;
	// geometric centre position (in microns with respect to the scene)
	i3d::Vector3d<float> sphereCentre;
	
	// used for visualization
	const size_t cellID;
		
};


/**
 * \ingroup toolbox
 *
 * Class handling locking and unlocking of custom defined locks.
 * Manages a list of current locks derived from the 'CustomLock' class.
 *
 */
class SceneLocks
{
public:

	
#ifdef GTGEN_WITH_LOCKVISUALIZATION
	// special constructor needed for visualization
	SceneLocks():
		startTime(report_st.GetStartTime()) {}
#endif
	
	/// a single frame lock (locks the entire frame)
	lock_handle lock(
		const size_t timePoint,
		const scene_type sceneType = all)
	{ return lock(timePoint, timePoint, sceneType); }
	
	/// interval of frames lock (locking interval is inclusive)
	lock_handle lock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const scene_type sceneType = all)
	{		
		CustomLock* newLock = new TimePointLock(fromTimePoint, toTimePoint, sceneType);	
		return checkAndLock(newLock);	
	}
	
	/// DEFAULT single frame Sphere-shaped lock
	template <class MV, class PV> 
	lock_handle lock(
		const size_t timePoint,
		const Cell<MV,PV>* const cellPtr,
		const scene_type sceneType = all)
	{ return lock(timePoint, timePoint, cellPtr, sceneType); }
	
	/// DEFAULT interval of frames Sphere-shaped lock (locking interval is inclusive)
	template <class MV, class PV> 
	lock_handle lock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const Cell<MV,PV>* const cellPtr,
		const scene_type sceneType = all)
	{
		CustomLock* newLock = new SphereLock(fromTimePoint, toTimePoint, cellPtr, sceneType);		
		return checkAndLock(newLock);
	}

	/// CUSTOM single frame Sphere-shaped lock
	lock_handle lock(
		const size_t timePoint,
		const SphereLock& SphLock)
	{ return lock(timePoint, timePoint, SphLock); }

	/// CUSTOM interval of frames Sphere-shaped lock (locking interval is inclusive)
	lock_handle lock(
		const size_t fromTimePoint,
		const size_t toTimePoint,
		const SphereLock& SphLock)
	{
		CustomLock* newLock = new SphereLock(fromTimePoint, toTimePoint, SphLock);
		return checkAndLock(newLock);
	}

	/// acquire the lock
	lock_handle checkAndLock(CustomLock* newLock);
		
	/// unlock given lock
	void unlock(lock_handle lockhnd)
	{
		//DEBUG_REPORT("*** Deleting custom lock.");

		std::unique_lock<std::mutex> listSearchLock(listSearchLock_st);
		
#ifndef GTGEN_WITH_LOCKVISUALIZATION
		// delete object
		delete (*lockhnd);
		
		// delete list entry
		listOfCustomLocks.erase(lockhnd);
#else
		// save oldLock pointer for visualization
		CustomLock* oldLock = *lockhnd;
		
		// delete list entry (so the lock won't be drawn in the upcomming visualization)
		listOfCustomLocks.erase(lockhnd);

		// save visualization
		saveVisualization(oldLock, false);
			
		// delete object
		delete oldLock;	
#endif
		
		listSearchLock.unlock();
		justUnlocked.notify_all();
	}
	
	/// unlock ALL custom locks created by the calling thread
	/// (used in case of exception)
	void unlock()
	{
		std::unique_lock<std::mutex> listSearchLock(listSearchLock_st);
		
		std::list<CustomLock*>::iterator iter = listOfCustomLocks.begin();
		
		while (iter != listOfCustomLocks.end())
		{
			if ((*iter)->getThreadID() == thrID)
			{
				// delete object
				delete (*iter);
		
				// delete list entry & move the iterator to the next element
				iter = listOfCustomLocks.erase(iter);
			}
			else
				++iter;
		}
		
		listSearchLock.unlock();
		DEBUG_THR_REPORT("Unlocked all locks of thread " << thrID << "!");
		justUnlocked.notify_all();
	}
	
	/// check whether given locks are intersecting
	bool ableToLockTimePoints(CustomLock* curLock, CustomLock* newLock);	
	bool ableToLockSphere(CustomLock* curLock, CustomLock* newLock);
	
#ifdef GTGEN_WITH_LOCKVISUALIZATION
	/// visualization methods
	void setImgRes(const i3d::Vector3d<float> res)
	{ imgRes = res; }
	
	void setImgSize(const i3d::Vector3d<size_t> size)
	{ imgSize = size; }
	
	void saveVisualization(CustomLock* const newLock, const bool addedLock) const;
#endif

private:

	std::list<CustomLock*> listOfCustomLocks;
	std::mutex listSearchLock_st;
	std::condition_variable justUnlocked;
	
#ifdef GTGEN_WITH_LOCKVISUALIZATION
	// visualization parameters
	i3d::Vector3d<float> imgRes;
	i3d::Vector3d<size_t> imgSize;
	const report_clock::time_point startTime;
#endif
		
};


#endif
